package fastaIndex;


import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class FastaFile extends LineOrientedFile {


	public FastaFile(File dir, String name) {
		super(dir,name);
	}
	
	public FastaFile(String path) {
		super(path);
	}
	
	public HashMap<String, FastaEntry> createHeaderIndex() throws IOException {
		return createHeaderIndex(new DefaultFastaHeaderParser('|', new String[0]),new HashMap<String, FastaEntry>());
	}
	
	
	public <M extends Map<String,FastaEntry>> M createHeaderIndex(FastaHeaderParser parser, M map) throws IOException {
		Iterator<FastaEntry> it = entryIterator(true);
		while (it.hasNext()) {
			FastaEntry e = it.next();
			String key = parser.getId(e.getHeader());
			map.put(key, e);
		}
		return map;
	}
	
	private static String getFullNameWithoutExtension(File f) {
		String path= f.getAbsolutePath();
		int dotIndex = path.lastIndexOf('.');
		int slashIndex = path.lastIndexOf(File.pathSeparatorChar);
		if (dotIndex>slashIndex)
			return path.substring(0,dotIndex);
		else
			return path;
	}
	
	public FastaIndexFile obtainDefaultIndex() {
		return new FastaIndexFile(getFullNameWithoutExtension(this)+".fi");
	}
	public FastaIndexFile obtainAndOpenDefaultIndex() throws IOException {
		FastaIndexFile fi = obtainDefaultIndex();
		if (fi.exists())
			return fi.open();
		else
			return fi.create(this);
	}
	
	@Override
	public void finishWriting() throws IOException {
		super.finishWriting();
	}
	public void writeEntry(FastaEntry fe) throws IOException {
		writeLine(fe.toString());
	}
	
	
	public Iterator<FastaEntry> entryIterator(boolean reUseEntry) throws IOException {
		return new EntryIterator(createReader(),reUseEntry);
	}
	
	public static class EntryIterator implements Iterator<FastaEntry> {

		private FastaEntry entry;
		private boolean reUse;
		
		private FastaEntry next;
		
		private BufferedReader br;
		private String header;
		
		public EntryIterator(Reader reader, boolean reUse) throws IOException {
			this.reUse = reUse;
			br = new BufferedReader(reader);
			header = br.readLine();
		}

		@Override
		public boolean hasNext() {
			lookAhead();
			return next!=null;
		}

		@Override
		public FastaEntry next() {
			lookAhead();
			FastaEntry re = next;
			next = null;
			return re;
		}

		private void lookAhead() {
			if (next==null && header!=null) {
				next = reUse && entry !=null ? entry : new FastaEntry();
				
				if (!header.startsWith(">")) {
					split(next,header);
					try {
						header = br.readLine();
					} catch (IOException e) {
						throw new RuntimeException("Could not read file!",e);
					}
				} else {
					StringBuilder seq = new StringBuilder();
					String line = null;
					try {
						while ((line=br.readLine())!=null && !line.startsWith(">"))
							seq.append(line);
						
					} catch (IOException e) {
						throw new RuntimeException("Could not read file!",e);
					}
					
					next.setHeader(header);
					next.setSequence(seq.toString());
					header = line;
				}
			}
		}
		
		private void split(FastaEntry e, String line) {
			int index = line.lastIndexOf(':');
			e.setHeader(index>=0 ? line.substring(0,index) : "");
			e.setSequence(line.substring(index+1));
		}

		@Override
		public void remove() {}
		
	}
	
	
	public static class FastaEntry {
		
		private String sequence;
		private String header;
		private int width = 80;

		public FastaEntry() {
		}
		
		public FastaEntry(String header, String sequence) {
			setHeader(header);
			setSequence(sequence);
		}

		public String getSequence() {
			return sequence;
		}

		public void setSequence(String sequence) {
			this.sequence = sequence;
		}

		/**
		 * Includes the >
		 * @return
		 */
		public String getHeader() {
			return header;
		}

		public void setHeader(String header) {
			this.header = header.startsWith(">")?header:">"+header;
		}
		
		public void setWidth(int width) {
			this.width = width;
		}
		
		public int getWidth() {
			return width;
		}
		
		@Override
		public String toString() {
			if (width<=0)
				return header+"\n"+sequence;
			else
				return header+"\n"+concat("\n", split(sequence,width));
		}
		
		
		public FastaEntry clone() {
			return new FastaEntry(header,sequence);
		}
		
	}
	
	public static interface FastaHeaderParser {

		public abstract boolean isReUseMap();
		public abstract void setReUseMap(boolean reUseMap);
		
		public abstract Map<String,String> parseHeader(String header);
		public abstract boolean canParse(String header);
		public abstract String getId(String header);
		
	}


	public static abstract class AbstractFastaHeaderParser implements FastaHeaderParser {

		private boolean reUseMap = true;
		private Map<String,String> map;
		private int numFields;
		
		public AbstractFastaHeaderParser(int numFields) {
			this.numFields = numFields;
		}
	
		public boolean isReUseMap() {
			return reUseMap;
		}
	
		public void setReUseMap(boolean reUseMap) {
			this.reUseMap = reUseMap;
		}
	
		@Override
		public boolean canParse(String header) {
			return numFields<0 || parseHeader(header).size()==numFields;
		}
		
		protected Map<String, String> getMap() {
			Map<String,String> re = 
				 reUseMap && map!=null ? map : new HashMap<String, String>();
			re.clear();
			return re;
		}
		
		
	}
	
	public class DefaultFastaHeaderParser extends AbstractFastaHeaderParser{

		private Pattern pattern;
		private String[] keys;
		private Character splitChar;
		
		
		public DefaultFastaHeaderParser() {
			super(1);
		}
		
		public DefaultFastaHeaderParser(Pattern pattern, String[] keys) {
			super(keys.length);
			this.pattern = pattern;
			this.keys = keys;
		}
		
		public DefaultFastaHeaderParser(Pattern pattern) {
			super(-1);
			this.pattern = pattern;
		}
		
		/**
		 * Segments specifies the start (inclusive) and stop (exclusive) position within the header
		 * @param segments
		 * @param keys
		 */
		public DefaultFastaHeaderParser(char splitChar, String[] keys) {
			super(keys.length);
			this.splitChar = splitChar;
			this.keys = keys;
		}
		
		/**
		 * Segments specifies the start (inclusive) and stop (exclusive) position within the header
		 * @param segments
		 * @param keys
		 */
		public DefaultFastaHeaderParser(char splitChar) {
			super(-1);
			this.splitChar = splitChar;
		}
		
		@Override
		public String getId(String header) {
			if (splitChar!=null) {
				String[] splitted = split(header.substring(1), splitChar);
				return splitted[0];
			} else if (pattern!=null){
				Matcher m = pattern.matcher(header);
				if (m.find())
					return m.group(1);
			}
			return header.substring(1);
		}

		@Override
		public Map<String, String> parseHeader(String header) {
			Map<String,String> re = getMap();

			if (splitChar!=null) {
				String[] splitted = split(header.substring(1), splitChar);
				for (int i=0; i<keys.length; i++) 
					re.put(keys==null?i+"":keys[i], splitted.length>=i?splitted[i]:null);
			} else if (pattern!=null) {
				Matcher m = pattern.matcher(header);
				if (m.find())
					for (int i=0; i<keys.length; i++) 
						if (keys==null || keys[i]!=null)
							re.put(keys==null?i+"":keys[i], m.group(i+1));
			} else {
				re.put("header", header);
			}
					
			return re;
		}
		
		
	}

	
	private static String concat(String glue,String[] a) {
		if (a.length==0) return "";
		StringBuilder re = new StringBuilder();
		re.append(a[0]);
		for (int i=1; i<a.length; i++) {
			re.append(glue);
			re.append(a[i]);
		}
		return re.toString();
	}
	
	private static String[] split(CharSequence s, int chunk) {
		String[] re = new String[(int)Math.ceil(s.length()/(double)chunk)];
		for (int i=0; i<re.length; i++)
			re[i] = s.subSequence(i*chunk, Math.min((i+1)*chunk,s.length())).toString();
		return re;
	}
	
	private static String[] split(String s, char separator) {
		if (s.length()==0)
			return new String[0];
		
		ArrayList<String> re = new ArrayList<String>();
		int p = 0;
		for (int sepIndex=s.indexOf(separator); sepIndex>=0; sepIndex = s.substring(p).indexOf(separator)) {
			re.add(s.substring(p,p+sepIndex));
			p += sepIndex+1;
		}
		re.add(s.substring(p));
		return re.toArray(new String[re.size()]);
	}
}
