package fastaIndex;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;



public class FastaIndexFile extends LineOrientedFile {

	private RandomAccessFile fastaFile;
	private HashMap<String,FastaIndexEntry> index;


	public FastaIndexFile(String path) {
		super(path);
	}

	public FastaIndexFile(File dir, String name) {
		super(dir, name);
	}

	public FastaIndexFile create(File fasta) throws IOException {
		return create(fasta,true);
	}
	

	public FastaIndexFile create(File fasta, boolean write) throws IOException {
		fastaFile = new RandomAccessFile(fasta, "r");
		index = new HashMap<String, FastaIndexFile.FastaIndexEntry>();
		long L = fastaFile.length();

		ArrayList<String> order = new ArrayList<String>();
		fastaFile.skipBytes(1);
		while (fastaFile.getFilePointer()<L) {
			String id = fastaFile.readLine();
			long start = fastaFile.getFilePointer();
			int lineWidth = fastaFile.readLine().length();
			long end = start;
			int first = fastaFile.read();
			while (fastaFile.getFilePointer()<L && first!='>') {
				if (first=='>') break;
				if (first=='\n'||first=='\r') {
					end=fastaFile.getFilePointer()-2;
					advanceToNextEntry(id,L);
					break;
				}
				if (skipLine()+1!=lineWidth) {
					end=fastaFile.getFilePointer()-1;
					advanceToNextEntry(id,L);
					break;
				}
				first = fastaFile.read();
				
			}
			if (end==start) 
				end=fastaFile.getFilePointer()-2;

			index.put(id,new FastaIndexEntry(id,start, end, lineWidth));
			order.add(id);
		}

		if (write) {
			startWriting();
			writeLine(fasta.getAbsoluteFile().getParentFile().equals(getAbsoluteFile().getParentFile())?fasta.getName():fasta.getAbsolutePath());
			for (String id : order) 
				writeLine(index.get(id).toString());
			finishWriting();
		}
		
		return this;
	}


	private int skipLine() throws IOException {
		boolean eol = false;
		RandomAccessFile fastaFile = this.fastaFile;
		int l = 0;
		while (!eol) {
			int r = fastaFile.read();
			switch (r) {
			case -1:
			case '\n':
				eol = true;
				break;
			case '\r':
				eol = true;
				long cur = fastaFile.getFilePointer();
				if ((fastaFile.read()) != '\n') {
					fastaFile.seek(cur);
				}
				break;
			default:
				l++;
				break;
			}
		}
		return l;
	}

	private void advanceToNextEntry(String id, long L) throws IOException {
		for (byte b=10; fastaFile.getFilePointer()<L && b!=62; b=fastaFile.readByte())
			if (b!=13 && b!=10)
				throw new IOException("Invalid fasta at "+id+" ("+(int)b+")!");
	}

	public FastaIndexFile open() throws IOException {
		if (fastaFile==null) {
			Iterator<String> it = lineIterator("#");
			fastaFile = new RandomAccessFile(new File(new File(getParentFile(),it.next()).getAbsolutePath()), "r");
			index = new HashMap<String, FastaIndexFile.FastaIndexEntry>();
			while (it.hasNext()) {
				String[] a = split(it.next(),'\t');
				index.put(a[0],new FastaIndexEntry(a[0],Long.parseLong(a[1]),Long.parseLong(a[2]),Integer.parseInt(a[3])));
			}
		}
		return this;
	}
	
	private static String[] split(String s, char separator) {
		if (s.length()==0)
			return new String[0];
		
		ArrayList<String> re = new ArrayList<String>();
		int p = 0;
		for (int sepIndex=s.indexOf(separator); sepIndex>=0; sepIndex = s.substring(p).indexOf(separator)) {
			re.add(s.substring(p,p+sepIndex));
			p += sepIndex+1;
		}
		re.add(s.substring(p));
		return re.toArray(new String[re.size()]);
	}

	public boolean isOpen() {
		return fastaFile!=null;
	}

	public void close() throws IOException {
		fastaFile.close();
		fastaFile = null;
		index = null;
	}

	public Set<String> getEntryNames() {
		return index.keySet();
	}

	public long getLength(String name) {
		return index.get(name).length();
	}

	/**
	 * start zerobased inclusive, end zerobased exclusive
	 * @param name
	 * @param start
	 * @param end
	 * @return
	 * @throws IOException
	 */
	public String getSequence(String name, long start, long end) throws IOException {
		FastaIndexEntry e = index.get(name);
		if (e==null)return null;
		return e.getSequence(start, end);
	}
	public String getSequence(String name, long start, long end, int flank) throws IOException {
		FastaIndexEntry e = index.get(name);
		if (e==null)return null;
		return e.getSequence(start, end,flank);
	}
	
	public String getSequence(String name) throws IOException {
		FastaIndexEntry e = index.get(name);
		if (e==null)
			return null;
		return e.getSequence(0, e.length());
	}
	
	public boolean containsEntry(String name) {
		return index.containsKey(name);
	}

	public FastaIndexEntry getEntry(String name) {
		return index.get(name);
	}

	public class FastaIndexEntry {
		private String id;
		private long start;
		private long end;
		private int lineWidth;
		private long length;
		public FastaIndexEntry(String id,long start, long end, int lineWidth) {
			this.id = id;
			this.start = start;
			this.end = end;
			this.lineWidth = lineWidth;
			this.length = end-start-(end-start)/(lineWidth+1);
		}
		public long length() {
			return length;
		}
		public String getSequence() throws IOException {
			return getSequence(0,length);
		}
		public String getSequence(long start, long end) throws IOException {
			return getSequence(start,end,0);
		}
		public String getSequence(long start, long end, int flank) throws IOException {
			start = Math.max(start, 0);
			end = Math.min(end,length);
			long fstart = Math.max(start-flank, 0);
			long fend = Math.min(end+flank,length);
			
			byte[] re = new byte[(int) (fend-fstart)];
			long rem = fstart%lineWidth;
			long bin = fstart+fstart/lineWidth;
			fastaFile.seek(this.start+bin);
			fastaFile.read(re, 0, Math.min((int) (lineWidth-rem),re.length));
			fastaFile.skipBytes(1);
			for (int i=(int) (lineWidth-rem); i<re.length; i+=lineWidth){
				fastaFile.read(re, i, Math.min(re.length-i,lineWidth));
				fastaFile.skipBytes(1);
			}
			if (flank>0) {
				String s = new String(re);
				s = s.substring(0,(int) (start-fstart)).toLowerCase()+
					s.substring((int) (start-fstart),(int) (end-fstart)).toUpperCase()+
					s.substring((int) (end-fstart)).toLowerCase();
				return s;
			}
			return new String(re);
		}
		public long getStart() {
			return start;
		}
		public long getEnd() {
			return end;
		}
		public int getLineWidth() {
			return lineWidth;
		}
		@Override
		public String toString() {
			return String.format("%s\t%d\t%d\t%d", id,start,end,lineWidth);
		}
		

	}


	
	public static void main(String[] args) throws IOException {
		if (args.length==0) {
			System.out.println("java FastaIndexFile <index-file>\nreads <chr>\\t<start>\\t<end> from stdin (zerobased, start inclusive, end exclusive)");
			System.exit(0);
		}
		
		FastaIndexFile f = new FastaIndexFile(args[0]);
		f.open();
		Iterator<String> it = new LineOrientedFile(LineOrientedFile.STDOUT).lineIterator();
		while (it.hasNext()) {
			String[] t = it.next().split("\t");
			System.out.println(f.getSequence(t[0], Long.parseLong(t[1]), Long.parseLong(t[2])));
		}
		f.close();
	}

}
