package fastaIndex;


import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.Locale;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;


public class LineOrientedFile extends File {

	public static final String STDOUT = "stdout";
	public static final String STDERR = "stderr";
	
	protected Writer fw;
	private Boolean gzipped;
	protected boolean autoFlush = false;
	
	public LineOrientedFile(File dir, String name) {
		super(dir,name);
		if (isPipe()) {
			gzipped = false;
			autoFlush = true;
		}
	}
	
	public LineOrientedFile(String path) {
		super(path);
		if (isPipe()) {
			gzipped = false;
			autoFlush = true;
		}
	}
	
	public boolean isGZIP() throws IOException {
		if (gzipped==null) {
			if (!exists())
				gzipped = false;
			else {
		        InputStream in = createInputStream();
		        try {
		            gzipped = in.read() == (GZIPInputStream.GZIP_MAGIC & 0xFF)
		                && in.read() == (GZIPInputStream.GZIP_MAGIC >> 8);
		        } finally {
		            in.close();
	        }
			}
		}
		return gzipped;
    }
	
	public long guessNumOfLines() throws IOException {
		return guessNumOfLines(100);
	}
	public long guessNumOfLines(int useFirstLines) throws IOException {
		long firstLinesLength = 0;
		Iterator<String> it = lineIterator();
		int i = 0;
		while (it.hasNext() && i++<useFirstLines)
			firstLinesLength+=it.next().length();
		if (!it.hasNext())
			return firstLinesLength;
		else
			return (long)((double)length()/((double)firstLinesLength/(double)useFirstLines));
	}
	
	public void setAutoFlush(boolean autoFlush) {
		this.autoFlush = autoFlush;
	}
	
	public boolean isAutoFlush() {
		return autoFlush;
	}
	
	public void startAppending() throws IOException {
		startWriting(false,true);
	}
	
	public boolean isWriting() {
		return fw!=null;
	}
	
	public Writer startWriting() throws IOException {
		return startWriting(isGZIP(),false);
	}
	
	public Writer startWriting(boolean gzip) throws IOException {
		return startWriting(gzip, false);
	}
	
	public Writer startWriting(boolean gzip, boolean appending) throws IOException {
		gzipped = gzip;
		if (getParentFile()!=null && !getParentFile().exists())
			getParentFile().mkdirs();
		fw = createWriter(appending);
		return fw;
	}
	
	public void writeLine(String line) throws IOException {
		fw.write(line);
		fw.write("\n");
		if (autoFlush)
			fw.flush();
	}
	
	public void write(String line) throws IOException {
		fw.write(line);
		if (autoFlush)
			fw.flush();
	}
	
	public void writef(String line, Object...args) throws IOException {
		fw.write(String.format(Locale.US,line,args));
		if (autoFlush)
			fw.flush();
	}
	
	public void finishWriting() throws IOException {
		if (!isPipe()) {
			fw.flush();
			fw.close();
		}
		fw = null;
	}
	
	
	
	public boolean isPipe() {
		return getPath().equals(STDOUT) || getPath().equals(STDERR);
	}
	
	boolean forbiddenLineIterator = false;
	private String enc = Charset.defaultCharset().name();
	
	public Iterator<String> lineIterator(String...commentPrefixes) throws IOException {
		if (forbiddenLineIterator)
			throw new IOException("Creating another line iterator is forbidden!");
		
		if (isPipe())
			forbiddenLineIterator = true;
		
		return new LineIterator(commentPrefixes);
	}
	
	public void setEncoding(String charset) {
		this.enc = charset;
	}
	
	public String getEncoding() {
		return enc;
	}
	
	/**
	 * Not buffered!
	 * @return
	 * @throws IOException
	 */
	public InputStreamReader createReader() throws IOException {
		return isGZIP() ?
				new InputStreamReader(new GZIPInputStream(createInputStream())) 
				: new InputStreamReader(createInputStream());
	}
	
	protected InputStream createInputStream() throws FileNotFoundException {
		if (isPipe())
			return System.in;
		else
			return new FileInputStream(this);
	}
	
	public Writer createWriter() throws IOException {
		return createWriter(false);
	}
	
	public Writer createWriter(boolean append) throws IOException {
		return isGZIP() ?
				new OutputStreamWriter(new GZIPOutputStream(createOutputStream(append)),enc) 
				: new OutputStreamWriter(createOutputStream(append),enc);
	}
	
	protected OutputStream createOutputStream(boolean append) throws FileNotFoundException  {
		if (this.getPath().equals(STDOUT))
			return System.out;
		else if (this.getPath().equals(STDERR))
			return System.err;
		else
			return new FileOutputStream(this,append);
	}
	
	protected void firstLineRead(String firstLine) {};
	
	protected class LineIterator implements Iterator<String> {

		public String nextLine = null;
		public BufferedReader br;
		public String[] commentPrefixes;
		public boolean closed = false;
		
		public LineIterator(String... commentPrefixes) throws IOException {
			this.commentPrefixes = commentPrefixes;
			br = new BufferedReader(createReader());
			lookAhead();
			firstLineRead(nextLine);
			
		}

		@Override
		public boolean hasNext() {
			lookAhead();
			return nextLine!=null;
		}

		@Override
		public String next() {
			lookAhead();
			String re = nextLine;
			nextLine=null;
			return re;
		}
		
		
		private void lookAhead() {
			if (nextLine==null) {
				if (closed)
					return;
				try {
					nextLine = br.readLine();
					if (commentPrefixes.length>0) 
						while (nextLine!=null && isComment(nextLine))
							nextLine = br.readLine();
					
					if (nextLine==null) {
						br.close();
						closed = true;
					}
				} catch (IOException e) {
					throw new RuntimeException("Could not iterate over lines!",e);
				}
			}
		}

		private boolean isComment(String line) {
			for (String cp : commentPrefixes)
				if (cp.length()==0 && line.length()==0)
					return true;
				else if (cp.length()>0 && nextLine.startsWith(cp)){
					return true;
				}
			return false;
		}
		
		@Override
		public void remove() {}
		
	}

	public String readAllText() throws IOException {
		StringBuilder sb = new StringBuilder();
		Iterator<String> it = lineIterator();
		while (it.hasNext()) {
			sb.append(it.next());
			sb.append("\n");
		}
		return sb.toString();
	}


}
